View subclass:#ListModelView
	instanceVariableNames:'list listHolder textStartLeft viewOrigin enabled fgColor bgColor
		lineSpacing widthOfContents computeWidthInRange startOfLinesY
		supportsDisplayInRange'
	classVariableNames:'DefaultForegroundColor DefaultBackgroundColor DefaultShadowColor
		DefaultLightColor StopRedrawSignal'
	poolDictionaries:''
	category:'AAA'
!

!ListModelView class methodsFor:'documentation'!

documentation
"
    This class can only passively display collections of elements.
    The class doesn't keep its own list, it works directly on
    the model( List or HierarchicalList ).

    Each list item is obligated to provide 3 services:
        - heightOn:aGC
        - widthOn:aGC
        - displayOn:aGC x:x y:y

    [Instance variables:]
        list                   <List-Model>     List or HierarchicalList ...
        listHolder             <Model>          Model, which keeps a List
        textStartLeft          <Number>         left inset of text
        viewOrigin             <Point>          the current origin
        enabled                <Boolean>        widget is enabled/disabeled
        fgColor                <Color>          color to draw characters
        bgColor                <Color>          the background
        lineSpacing            <Number>         pixels between lines
        widthOfContents        <Number>         cached width of widest line
        computeWidthInRange    <Point>          used for recompute width of contents
        startOfLinesY          <Collection>     keeps all the absolute Y-start positions
                                                for each line in the list. The first
                                                entry into the list is the top Y inset.
        supportsDisplayInRange <Boolean>        set to true if the list elements
                                                supports the service:
                                                    #displayOn:x:y:toX:y:

    [author:]
        Claus Atzkern

    [see also:]

        SelectionInListModelView
        HierarchicalListView
        List
        HierarchicalList
"



!

examples
"
                                                                        [exBegin]
    |top list view up index|

    list := List new.

    top  := StandardSystemView new; extent:300@300.
    view := ScrollableView for:ListModelView miniScroller:true
                        origin:0.0@0.0 corner:1.0@1.0 in:top.
    view list:list.
    top  openAndWait.
    up := true.

    [top shown] whileTrue:[
        Delay waitForSeconds:0.2.

        up ifTrue:[
            index := 1 + (list size).
            list add:('element: ', index printString).
            up := index < 10
        ] ifFalse:[
            list removeIndex:1.
            up := list isEmpty.
        ]
    ].
                                                                        [exEnd]

"

! !

!ListModelView class methodsFor:'initialization'!

initialize
    "setup the signals"

    StopRedrawSignal := ErrorSignal newSignalMayProceed:false.


! !

!ListModelView class methodsFor:'Signal constants'!

stopRedrawSignal
    ^ StopRedrawSignal
! !

!ListModelView class methodsFor:'defaults'!

updateStyleCache
    "extract values from the styleSheet and cache them in class variables"

    <resource: #style   (
                        #'text.backgroundColor'  #'text.foregroundColor'
                        #'selection.shadowColor' #'selection.lightColor'
                        #'text.font'
                        )>

    DefaultForegroundColor         := StyleSheet colorAt:'text.foregroundColor' default:Black.
    DefaultBackgroundColor         := StyleSheet colorAt:'text.backgroundColor'.
    DefaultShadowColor             := StyleSheet colorAt:'selection.shadowColor'.
    DefaultLightColor              := StyleSheet colorAt:'selection.lightColor'.
    DefaultFont                    := StyleSheet  fontAt:'text.font'.

    "
     self updateStyleCache
    "


! !

!ListModelView methodsFor:'accessing'!

list
    "get the list of items
    "
    ^ list


!

list:aList
    "set the list of items
    "
    |size changed negatedOrg|

    changed := false.

    list notNil ifTrue:[
        changed := list notEmpty.
        list removeDependent:self.
    ].
    list := aList.
    self recomputeHeightOfContents.

    list notNil ifTrue:[
        changed := list notEmpty.
        list addDependent:self.
    ].
    preferredExtent := nil.
    widthOfContents := nil.

    viewOrigin = (0@0) ifFalse:[
        self originWillChange.
        negatedOrg := viewOrigin negated.
        viewOrigin := (0@0).
        self originChanged:negatedOrg.
    ]

    changed ifTrue:[
        self invalidate.
        self contentsChanged
    ]

!

size
    "returns number of raws
    "
    ^ list size


! !

!ListModelView methodsFor:'accessing behavior'!

enabled
   "returns the enabled state
   "
   ^ enabled


!

enabled:aBoolean
   "set the enabled state
   "
   enabled := aBoolean


! !

!ListModelView methodsFor:'accessing items'!

at:anIndex
    "return the list element at an index
    "
    ^ list at:anIndex
!

at:anIndex ifAbsent:exceptionBlock
    "return the list element at an index if valid. 
     If the index is invalid, return the result of evaluating 
     the exceptionblock.
    "
    list notNil ifTrue:[
        ^ list at:anIndex ifAbsent:exceptionBlock
    ].
    ^ nil.
!

identityIndexOf:anElement
    "returns the index of an element or nil
    "
    ^ list notNil ifTrue:[list identityIndexOf:anElement]
                 ifFalse:[0]


! !

!ListModelView methodsFor:'accessing look'!

backgroundColor
    "get the background color
    "
    ^ bgColor


!

backgroundColor:aColor
    "set the background color
    "
    (aColor notNil and:[bgColor ~~ aColor]) ifTrue:[
        super viewBackground:bgColor.

        self realized ifTrue:[
            bgColor := aColor on:device.
            self invalidate
        ] ifFalse:[
            bgColor := aColor
        ]
    ]

!

font:aFont
    "set a new font
    "
    |oldWidth oldHeight|

    (aFont isNil or:[aFont = font]) ifFalse:[
        oldWidth  := font width.
        oldHeight := font height.

        super font:aFont.

        (font width) ~~ oldWidth ifTrue:[       "/ force a recomputation
            preferredExtent := nil.
            widthOfContents := nil.
        ].
        oldHeight ~~ font height ifTrue:[
            self recomputeHeightOfContents.
        ].
        shown ifTrue:[ self invalidate ]
    ].
!

foregroundColor
    "get the foreground color
    "
    ^ fgColor

!

foregroundColor:aColor
    "set the foreground color
    "
    (aColor notNil and:[fgColor ~~ aColor]) ifTrue:[
        self realized ifTrue:[
            fgColor := aColor on:device.
            self invalidate
        ] ifFalse:[
            fgColor := aColor
        ]
    ]

! !

!ListModelView methodsFor:'accessing mvc'!

listHolder
    "returns the listHolder or nil
    "
    ^ listHolder
!

listHolder:aListHolder
    "set a new listHolder
    "
    |newList|

    listHolder removeDependent:self.

    (listHolder := aListHolder) notNil ifTrue:[
        listHolder addDependent:self.
    ].
    self list:(listHolder value).
! !

!ListModelView methodsFor:'change & update'!

lineChangedAt:aLnNr with:arg
    "line changed at position; check whether line height changed
    "
    |
     oldHeight "{ Class:SmallInteger }"
     dltHeight "{ Class:SmallInteger }"
    |


    (arg notNil and:[(arg == #icon or:[arg == #hierarchy])]) ifTrue:[
        ^ self
    ].
    oldHeight := (self yVisibleOfLine:(aLnNr + 1)) - (self yVisibleOfLine:aLnNr).
    dltHeight := (self heightOfLineAt:aLnNr) - oldHeight.

    dltHeight == 0 ifTrue:[
        ^ self redrawLineAt:aLnNr
    ].

    aLnNr + 1 to:startOfLinesY size do:[:i|
        startOfLinesY at:i put:((startOfLinesY at:i) + dltHeight)
    ].
    self invalidate.


!

listChangedInsert:start nItems:nLines
    "list changed; items are added
    "
    |
     run    "{ Class: SmallInteger }"
     cpyHg  "{ Class: SmallInteger }"
     absY0  "{ Class: SmallInteger }"
     absY1  "{ Class: SmallInteger }"
     visY0  "{ Class: SmallInteger }"
     visY1  "{ Class: SmallInteger }"
     orgY   "{ Class: SmallInteger }"
     dltY   "{ Class: SmallInteger }"
     maxHg  "{ Class: SmallInteger }"
   |

    nLines == 0 ifTrue:[^ self ].
    self listSizeChanged:start nLines:nLines.
    shown not ifTrue:[^ self].

    startOfLinesY addAll:(Array new:nLines) beforeIndex:start + 1.
    absY0 := startOfLinesY at:start.
    absY1 := absY0.
    run   := start.

    nLines timesRepeat:[
        absY1 := absY1 + (self heightOfLineAt:run).
        run   := run + 1.
        startOfLinesY at:run put:absY1.
    ].
        
    dltY := absY1 - absY0.
    run + 1 to:(startOfLinesY size) do:[:i|
        startOfLinesY at:i put:((startOfLinesY at:i) + dltY)
    ].

    orgY  := viewOrigin y.
    absY1 := absY0 + dltY.
    visY0 := absY0 - orgY.
    visY1 := absY1 - orgY.
    maxHg := height - margin.

    absY0 < orgY ifTrue:[
        self originWillChange.
        viewOrigin y:(dltY + orgY).
        self originChanged:(0 @ dltY).
    ].

    (self sensor hasDamageFor:self) ifTrue:[
        ^ self contentsChanged.
    ].

    (visY0 >= maxHg or:[visY1 <= margin]) ifTrue:[
        ^ self contentsChanged
    ].
    visY0 := visY0 max:margin.
    visY1 := visY1 min:maxHg.

    (start == list size or:[(cpyHg := maxHg - visY1) < 20]) ifTrue:[
        visY1 := maxHg
    ] ifFalse:[
        self catchExpose.

        self copyFrom:self x:0 y:visY0
                         toX:0 y:visY1
                       width:width height:cpyHg async:true.

        self waitForExpose
    ].

    self redrawX:margin y:visY0
                    width:width - margin - margin
                   height:(visY1 - visY0).
    self contentsChanged.


!

listChangedRemove:start toIndex:stop
    "list changed; items are removed
    "
    |noRedraw
     size  "{ Class: SmallInteger }"
     absY0 "{ Class: SmallInteger }"
     absY1 "{ Class: SmallInteger }"
     visY0 "{ Class: SmallInteger }"
     visY1 "{ Class: SmallInteger }"
     orgY  "{ Class: SmallInteger }"
     orgX  "{ Class: SmallInteger }"
     dltY  "{ Class: SmallInteger }"
     cpyHg "{ Class: SmallInteger }"
     maxHg "{ Class: SmallInteger }"
    |

    (size := stop - start + 1) == 0 ifTrue:[
        ^ self
    ].
    self listSizeChanged:start nLines:(size negated).
    shown not ifTrue:[^ self].
    absY0 := self yAbsoluteOfLine:start.
    absY1 := self yAbsoluteOfLine:(stop + 1).
    dltY  := absY1 - absY0.

    startOfLinesY removeFromIndex:start toIndex:stop.

    start to:(startOfLinesY size) do:[:i|
        absY0 := startOfLinesY at:i.
        startOfLinesY at:i put:(absY0 - dltY).
    ].

    orgY  := viewOrigin y.
    orgX  := viewOrigin x.

    absY0 := self yAbsoluteOfLine:start.
    absY1 := absY0 + dltY.
    visY0 := absY0 - orgY.
    visY1 := absY1 - orgY.
    maxHg := height - margin.

    (list size == 0 or:[(orgY ~~ 0 and:[self maxViewOriginY == 0])]) ifTrue:[
        (orgX ~~ 0 or:[orgY ~~ 0]) ifTrue:[
            self originWillChange.
            viewOrigin := 0@0.
            self originChanged:((orgX @ orgY) negated).
        ].
        self invalidate.
      ^ self contentsChanged    
    ].

    visY0 < margin ifTrue:[
        noRedraw := visY1 <= margin.
        noRedraw ifTrue:[dltY := dltY negated] ifFalse:[dltY := visY0].

        self originWillChange.
        viewOrigin y:(dltY + orgY).
        self originChanged:(0 @ dltY).        
    ] ifFalse:[
        noRedraw := visY0 >= maxHg
    ].

    (noRedraw or:[self sensor hasDamageFor:self]) ifFalse:[
        visY0 := visY0 max:margin.
        cpyHg := maxHg - visY1.

        cpyHg > 20 ifTrue:[
            self catchExpose.
            self copyFrom:self x:0 y:visY1 toX:0 y:visY0
                    width:width height:cpyHg async:true.
            self waitForExpose.
            visY0 := visY0 + cpyHg.
        ].

        self redrawX:margin y:visY0
                        width:width - margin - margin
                       height:(maxHg - visY0).
    ].
    self contentsChanged.






!

update:what with:aPara from:chgObj
    "catch and handle any notification raised from the list model
     or list holder
    "
    |arg1 arg2|

    chgObj ~~ list ifTrue:[
        chgObj == listHolder ifTrue:[self  list:(listHolder value)]
                            ifFalse:[super update:what with:aPara from:chgObj].
      ^ self
    ].

    aPara isCollection ifFalse:[
        what == #at:     ifTrue:[ ^ self lineChangedAt:aPara with:nil].
        what == #insert: ifTrue:[ ^ self listChangedInsert:aPara nItems:1 ].
        what == #remove: ifTrue:[ ^ self listChangedRemove:aPara toIndex:aPara ].

      ^ self
    ].

    arg1 := aPara at:1.
    arg2 := aPara at:2.

    what == #at:               ifTrue:[ ^ self lineChangedAt:arg1 with:arg2 ].
    what == #insertCollection: ifTrue:[ ^ self listChangedInsert:arg1 nItems:arg2 ].
    what == #removeFrom:       ifTrue:[ ^ self listChangedRemove:arg1 toIndex:arg2 ].

    what == #replace: ifTrue:[
        arg1 to:arg2 do:[:i| self lineChangedAt:i with:nil ]
    ]





! !

!ListModelView methodsFor:'drawing'!

displayElement:anItem x:xL y:yT toX:xR y:yB
    "draw a label at xL/yT; fg/bg colors are already set
    "
    |ok
     h  "{ Class: SmallInteger }"
     y  "{ Class: SmallInteger }"
    |
    supportsDisplayInRange isNil ifTrue:[
        supportsDisplayInRange := anItem respondsTo:#displayOn:x:y:toX:y:
    ].
    supportsDisplayInRange ifTrue:[
        anItem displayOn:self x:xL y:yT toX:xR y:yB
    ] ifFalse:[
        h := yB - yT.
        y := yT - ((anItem heightOn:self) + 1 - h // 2).

        anItem isImageOrForm ifFalse:[
            (anItem isString or:[anItem isKindOf:LabelAndIcon]) ifTrue:[
                y := y + font ascent
            ]
        ].
        anItem displayOn:self x:xL y:y
    ]


!

invalidateX:x y:y width:w height:h
    "add a damage to redraw part of the view
    "
    shown ifTrue:[
        self invalidate:(Rectangle left:x top:y width:w height:h) 
              repairNow:false
    ]
!

redraw
    "redraw complete view
    "
    self redrawX:0 y:0 width:width height:height.


!

redrawLineAt:aLineNr
    "redraw a specific line
    "
    |y0 "{ Class:SmallInteger }"
     y1 "{ Class:SmallInteger }"
     yB "{ Class:SmallInteger }"
    |

    (shown and:[aLineNr notNil]) ifTrue:[
        yB := height - margin.
        y0 := (self yVisibleOfLine:aLineNr) max:margin.

        y0 < yB ifTrue:[
            y1 := (self yVisibleOfLine:(aLineNr + 1)) min:yB.
            y1 > margin ifTrue:[
                self redrawX:0 y:y0 width:width height:(y1 - y0)
            ]
        ]
    ]

!

redrawX:x y:y width:w height:h
    "redraw part of myself immediately, given logical coordinates 
    "
    |savClip start stop yAbs|

    shown ifFalse:[^ self].

    self paint:bgColor.
    self fillRectangleX:x y:y width:w height:h.

    (start := self yVisibleToLineNr:y) isNil ifTrue:[
        ^ self
    ].
    yAbs := y + h.

    (stop := self yVisibleToLineNr:yAbs) isNil ifTrue:[
        stop := list size.
    ] ifFalse:[
        yAbs == (startOfLinesY at:stop) ifTrue:[
            stop := stop - 1
        ].
    ].    

    savClip := clipRect.
    self clippingRectangle:(Rectangle left:x top:y width:w height:h).

    StopRedrawSignal handle:[:ex|  self invalidate.
                                   ex return
                            ] do:[ self drawFrom:start
                                              to:stop 
                                               x:x
                                               y:(self yVisibleOfLine:start)
                                             toX:x + w
                            ].

    self clippingRectangle:savClip.




! !

!ListModelView methodsFor:'event handling'!

keyPress:aKey x:x y:y
    "a key was pressed - handle page-keys here
    "
    <resource: #keyboard( #PreviousPage #NextPage  #HalfPageUp #HalfPageDown
                          #BeginOfText  #EndOfText #ScrollUp   #ScrollDown
                        )>

    |n sensor|

    list size ~~ 0 ifTrue:[
        (aKey == #PreviousPage) ifTrue:[^ self pageUp].
        (aKey == #NextPage)     ifTrue:[^ self pageDown].
        (aKey == #HalfPageUp)   ifTrue:[^ self halfPageUp].
        (aKey == #HalfPageDown) ifTrue:[^ self halfPageDown].
        (aKey == #BeginOfText)  ifTrue:[^ self scrollToTop].
        (aKey == #EndOfText)    ifTrue:[^ self scrollToBottom].

        (aKey == #ScrollUp or:[aKey == #ScrollDown]) ifTrue:[
            sensor := self sensor.

            n := sensor notNil ifTrue:[1 + (sensor compressKeyPressEventsWithKey:aKey)]
                              ifFalse:[1].

            n := n * self verticalScrollStep.

            aKey == #ScrollUp ifTrue:[self scrollUp:n]
                             ifFalse:[self scrollDown:n].
            ^ self
        ].
    ].
    super keyPress:aKey x:x y:y
!

wantsFocusWithButtonPress
    "catch the keyboard focus on button click
    "
    ^  true


! !

!ListModelView methodsFor:'fetch resources'!

colorOnDevice:aColor
    "fetch a device colors
    "
    ^ aColor notNil ifTrue:[aColor on:device] ifFalse:[nil]

!

imageOnDevice:anImage
    "associate image to device and clear pixel mask
     returns the new image.
    "
    |image|

    (image := anImage) notNil ifTrue:[
        image device ~~ device ifTrue:[
            image := image copy.
        ].
        image := image on:device.
        image := image clearMaskedPixels.
    ].
    ^ image

! !

!ListModelView methodsFor:'initialize / release'!

create
    "fetch device dependent resources
    "
    super create.
    self fetchResources.
    self recomputeHeightOfContents.
!

destroy
    "remove dependencies
    "
    listHolder removeDependent:self.
    list       removeDependent:self.
    super destroy

!

initStyle
    "setup viewStyle specifics
    "
    super initStyle.

    DefaultBackgroundColor notNil ifTrue:[
        viewBackground := DefaultBackgroundColor
    ].
    lineSpacing   := 0.
    textStartLeft := 2.
    fgColor       := DefaultForegroundColor.
    bgColor       := viewBackground.
    startOfLinesY := OrderedCollection new.

    startOfLinesY add:(2 + margin). "/ top inset of first line

    DefaultShadowColor notNil ifTrue:[
        shadowColor := DefaultShadowColor
    ].

    DefaultLightColor notNil ifTrue:[
        lightColor := DefaultLightColor
    ].
!

initialize
    "setup default attributes
    "
    super initialize.

    viewOrigin    := 0@0.
    bitGravity    := #NorthWest.
    enabled       := true.
!

recreate
    "fetch device dependent resources
    "
    super recreate.
    self fetchResources.

! !

!ListModelView methodsFor:'private'!

yAbsoluteOfLine:aLineNr
    "given a lineNr, return y-coordinate absolute
    "
    ^ startOfLinesY at:aLineNr ifAbsent:[startOfLinesY last].
!

yVisibleOfLine:aLineNr
    "given a lineNr, return y-coordinate in view
    "
    |y|

    y := startOfLinesY at:aLineNr ifAbsent:[startOfLinesY last].
  ^ y - viewOrigin y


!

yVisibleToLineNr:yVisible
    "returns the line number assigned to a physical y or nil
    "
    |
     yAbs    "{ Class: SmallInteger}"
     yMid    "{ Class: SmallInteger}"
     size    "{ Class: SmallInteger}"
     next    "{ Class: SmallInteger}"
     low     "{ Class: SmallInteger}"
     high    "{ Class: SmallInteger}"
     middle  "{ Class: SmallInteger}"
    |

    (size := startOfLinesY size) == 1 ifTrue:[^ nil].  "/ EMPTY LIST
    yAbs := yVisible + viewOrigin y.
    yAbs > (startOfLinesY at:size)    ifTrue:[^ nil].  "/ END OF LIST

    middle := size.
    size   := size - 1.
    high   := size.
    low    := 1.

    [(next := low + high // 2) ~~ middle] whileTrue:[
        yMid := startOfLinesY at:(middle := next).

        yMid < yAbs ifTrue:[low  := middle]
                   ifFalse:[high := middle]
    ].

    yAbs < yMid ifTrue:[
        ^ (middle - 1) max:1.
    ].
    middle < size ifTrue:[
        next := middle + 1.
        ^ (startOfLinesY at:next) > yAbs ifTrue:[middle] ifFalse:[next]
    ].
    ^ size
! !

!ListModelView methodsFor:'protocol'!

drawFrom:start to:stop x:xL y:yT toX:xR
    "draw lines from start to stop.
     clipping and clearing the background is already done
    "
    |item
     y0 "{ Class:SmallInteger }"
     y1 "{ Class:SmallInteger }"
     x0 "{ Class:SmallInteger }"
    |
    self paint:fgColor on:bgColor.

    x0 := textStartLeft - viewOrigin x.
    y1 := yT.

    start to:stop do:[:i|
        y0 := y1.
        y1 := self yVisibleOfLine:(i + 1).

        (item := list at:i ifAbsent:nil) notNil ifTrue:[
            self displayElement:item x:x0 y:y0 toX:xR y:y1
        ]
    ]

!

fetchResources
    "fetch device colors and ..., to avoid reallocation at redraw time;
     *** called after a create or snapin to fetch all device resources
    "
    fgColor    := self colorOnDevice:fgColor.
    bgColor    := self colorOnDevice:bgColor.
    font       := font on:device.

!

heightOfLineAt:aLineNr
    "returns the total height for a line at an index( including lineSpacing ... )
    "
    |item|

    item := list at:aLineNr ifAbsent:[^ 4].
  ^ lineSpacing + (item heightOn:self)
!

listSizeChanged:anIndex nLines:noLines
    "list size changed; information is stored to recompute the
     width if required( preferredExtent, horizontal scroller ... ).
     see: widthOfContents
     *** if nLines is negative, lines are removed otherwise added.
    "
    |start "{ Class:SmallInteger }"
     stop  "{ Class:SmallInteger }"
     size  "{ Class:SmallInteger }"
    |
    preferredExtent := nil.

    widthOfContents isNil ifTrue:[      "/ recompute whole list
        ^ self
    ].

    (noLines < 0 or:[(size := list size) <= noLines]) ifTrue:[
        widthOfContents := nil.         "/ force recompute whole list
        ^ self
    ].

    stop  := anIndex + noLines - 1.     "/ recompute a range
    start := anIndex.

    computeWidthInRange notNil ifTrue:[
        start := computeWidthInRange y.
        stop  <  start ifTrue:[stop := start min:size].
        start := (computeWidthInRange x) min:anIndex.
    ].
    computeWidthInRange := start@stop
!

widthOfWidestLineBetween:firstLine and:lastLine
    "return the width in pixels of the widest line in a range
    "
    |lbl item
     width    "{ Class: SmallInteger }"
    |
    width := textStartLeft.

    firstLine to:lastLine do:[:anIndex|
        item  := list at:anIndex ifAbsent:[^ width + textStartLeft ].
        width := (item widthOn:self) max:width
    ].
    ^ width + textStartLeft




! !

!ListModelView methodsFor:'recomputation'!

preferredExtent
    "returns the preferred extent
    "
    |x y|

    preferredExtent isNil ifTrue:[
        y := self heightOfContents.
        x := self widthOfContents.
        preferredExtent := x@y
    ].
  ^ preferredExtent


!

recomputeHeightOfContents
    "recompute all the y positions
    "
    |yAbs "{ Class: SmallInteger }"
     lnHg "{ Class: SmallInteger }"
    |

    yAbs := startOfLinesY at:1.   "/ top inset of first line
    startOfLinesY removeAll.
    startOfLinesY add:yAbs.

    1 to:(list size) do:[:anIndex|
        lnHg := self heightOfLineAt:anIndex.
        startOfLinesY add:(yAbs := yAbs + lnHg)
    ].
! !

!ListModelView methodsFor:'scroller interface'!

heightOfContents
    "return the height of the contents in pixels
    "
    ^ startOfLinesY last
!

innerHeight
    "returns the inner height of the contents shown
    "
    ^ height - margin - margin

!

innerWidth
    "returns the inner width of the contents shown
    "
    ^ width - margin - margin

!

maxViewOriginY
    "returns the maximum possible y of the view origin
    "
    ^ (self heightOfContents - self innerHeight) max:0


!

verticalScrollStep
    "return the amount to scroll when stepping up/down.
    "
    ^ 10

!

viewOrigin
    "return the viewOrigin; thats the coordinate of the contents 
     which is shown topLeft in the view.
    "
    ^ viewOrigin

!

widthOfContents
    "return the width of the contents in pixels
    "
    |range
     size  "{ Class:SmallInteger }"
     start "{ Class:SmallInteger }"
     stop  "{ Class:SmallInteger }"
    |
    (widthOfContents notNil and:[computeWidthInRange isNil]) ifTrue:[
        ^ widthOfContents + textStartLeft
    ].
    range               := computeWidthInRange.
    computeWidthInRange := nil.
    preferredExtent     := nil.
    size                := list size.

    size == 0 ifTrue:[
        widthOfContents := 20.
    ] ifFalse:[
        widthOfContents isNil ifTrue:[
            widthOfContents := self widthOfWidestLineBetween:1 and:size
        ] ifFalse:[
            start := range x.
            stop  := range y min:size.

            start > stop ifFalse:[
                size := self widthOfWidestLineBetween:start and:stop.

                widthOfContents < size ifTrue:[
                    widthOfContents := size
                ]
            ]
        ]
    ].
    ^ widthOfContents + textStartLeft
!

xOriginOfContents
    "return the horizontal origin of the contents in pixels
    "
    ^ viewOrigin x 

!

yOriginOfContents
    "return the vertical origin of the contents in pixels
    "
    ^ viewOrigin y

! !

!ListModelView methodsFor:'scrolling'!

halfPageDown
    "scroll down half a page
    "
    self scrollDown:(width // 2).

!

halfPageUp
    "scroll up half a page
    "
    self scrollUp:(width // 2).

!

scrollTo:anOrigin redraw:doRedraw
    "change origin to have newOrigin be visible at the top-left.
    "
    |newOrg dltOrg sensor
     innerWT    "{ Class:SmallInteger }"
     innerHG    "{ Class:SmallInteger }"
     h          "{ Class:SmallInteger }"
     x          "{ Class:SmallInteger }"
     x0         "{ Class:SmallInteger }"
     x1         "{ Class:SmallInteger }"
     y          "{ Class:SmallInteger }"
     w          "{ Class:SmallInteger }"
     y0         "{ Class:SmallInteger }"
     y1         "{ Class:SmallInteger }"
     dX         "{ Class:SmallInteger }"
     dY         "{ Class:SmallInteger }"
    |
    (shown and:[(sensor := self sensor) notNil]) ifFalse:[
        ^ self
    ].
    self windowGroup processExposeEvents.

    innerWT := self innerWidth.
    innerHG := self innerHeight.

    h := viewOrigin y.

    (y := anOrigin y) > h ifTrue:[              "/ end of contents
        y > (dY := self maxViewOriginY) ifTrue:[
            y := dY max:h
        ]
    ] ifFalse:[
        y := y max:0.
    ].

    (x := anOrigin x) > 0 ifTrue:[
        x := x min:(self widthOfContents - innerWT).
    ].
    x      := x max:0.
    newOrg := (x @ y).
    dltOrg := newOrg - viewOrigin.
    dX     := dltOrg x.
    dY     := dltOrg y.

    (dX == 0 and:[dY == 0]) ifTrue:[
        ^ self
    ].
    self originWillChange.
    viewOrigin := newOrg.

    doRedraw ifFalse:[
        ^ self originChanged:dltOrg
    ].

    dY ~~ 0 ifTrue:[                            "/ SCROLL VERTICAL
        dY := dY abs.

        (dX ~~ 0 or:[innerHG - dY < 20]) ifTrue:[
            self invalidate.
        ] ifFalse:[                             "/ COPY VERTICAL
            y0 := y1 := margin + dY.
            h  := innerHG - dY.

            dltOrg y < 0 ifTrue:[y0 := margin. y := y0]
                        ifFalse:[y1 := margin. y := y1 + h].

            self catchExpose.
            self copyFrom:self x:margin y:y0 toX:margin y:y1 width:innerWT height:h async:true.
            self waitForExpose.
            self invalidateX:margin y:y width:innerWT height:(innerHG - h).
        ]
    ] ifFalse:[                                 "/ SCROLL HORIZONTAL
        dX := dX abs.

        innerWT - dX < 20 ifTrue:[
            self invalidate.
        ] ifFalse:[                             "/ COPY HORIZONTAL
            x0 := x1 := dX + margin.
            w  := innerWT - dX.

            dltOrg x < 0 ifTrue:[x0 := x := margin ]
                        ifFalse:[x1 := margin. x := w].

            self catchExpose.
            self copyFrom:self x:x0 y:margin toX:x1 y:margin width:w height:innerHG async:true.
            self waitForExpose.
            self invalidateX:x y:margin width:(width - w) height:innerHG.
        ]
    ].
    self originChanged:dltOrg.
!

scrollToLine:aLineNumber
    "make line visible
    "
    |y  "{ Class:SmallInteger }"
     l  "{ Class:SmallInteger }"
    |
    (aLineNumber notNil and:[aLineNumber between:1 and:(list size)]) ifFalse:[
        ^ self
    ].

    y := self yVisibleOfLine:aLineNumber.

    y < margin ifTrue:[
        y := margin - y.
    ] ifFalse:[
        y := self yVisibleOfLine:(1 + aLineNumber).
        l := height - margin.
        y > l ifFalse:[^ self].
        y := l - y.
    ].
    self scrollTo:(viewOrigin - (0 @ y)).


! !

!ListModelView class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libwidg2/ListModelView.st,v 1.4 1999-05-28 18:07:15 cg Exp $'
! !
ListModelView initialize!
